/*
 * $Id: arpscan.c,v 1.2 2003/08/18 20:54:25 jason Exp $
 *
 * Copyright (c) 2002 Jason Ish <jason@codemonkey.net>
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <pcap.h>
#include <dnet.h>

#define PCAP_TIMEO 20

struct	ether_arp {
	struct	 arp_hdr ea_hdr;
	u_int8_t arp_sha[ETH_ADDR_LEN];
	u_int8_t arp_spa[4];
	u_int8_t arp_tha[ETH_ADDR_LEN];
	u_int8_t arp_tpa[4];
};

extern char	*__progname;
struct eth_hdr	 eth_header;
u_int32_t	 local_ip;

void
usage()
{
	fprintf(stderr, "USAGE: %s <ip-address>\n\n", __progname);
	fprintf(stderr, "\teg)\n");
	fprintf(stderr, "\t    %s 172.16.1.11\n", __progname);
	fprintf(stderr, "\t    %s 192.168.1.100-192.168.1.200\n", __progname);
	fprintf(stderr, "\t    %s 172.16.1.41/29\n");
	fprintf(stderr, "\n");
	exit(1);
}

void
pcap_cb(u_char *u, struct pcap_pkthdr *hdr, u_char *pkt)
{
	struct ether_arp *ether_arp;
	u_char eth_str[19];
	u_char ip_str[17];

	ether_arp = (struct ether_arp *)(pkt + sizeof(struct eth_hdr));

	if (ether_arp->ea_hdr.ar_op != ntohs(ARP_OP_REPLY))
		return;

	snprintf(eth_str, sizeof(eth_str) - 1, "%02x:%02x:%02x:%02x:%02x:%02x",
	    ether_arp->arp_sha[0], ether_arp->arp_sha[1],
	    ether_arp->arp_sha[2], ether_arp->arp_sha[3],
	    ether_arp->arp_sha[4], ether_arp->arp_sha[5]);

	snprintf(ip_str, sizeof(ip_str) - 1, "%d.%d.%d.%d",
	    ether_arp->arp_spa[0], ether_arp->arp_spa[1],
	    ether_arp->arp_spa[2], ether_arp->arp_spa[3]);

	printf("%-15s is at %s\n", ip_str, eth_str);
}

pcap_t *
init_pcap(char **dev_name)
{
	struct bpf_program bpf;
	char pcap_errbuf[PCAP_ERRBUF_SIZE];
	pcap_t *pcap;
	u_int32_t mask = 0xffffffff;

	if (*dev_name == NULL) {
		*dev_name = pcap_lookupdev(pcap_errbuf);
		if (*dev_name == NULL) {
			fprintf(stderr, "error: %s\n", pcap_errbuf);
			return (NULL);
		}
	}

	pcap = pcap_open_live(*dev_name, 64, 1, PCAP_TIMEO, pcap_errbuf);
	if (pcap == NULL) {
		fprintf(stderr, "error: %s\n", pcap_errbuf);
		return (NULL);
	}

	if (pcap_compile(pcap, &bpf, "arp", 1, 0) < 0) {
		pcap_close(pcap);
		return (NULL);
	}
	
	if (pcap_setfilter(pcap, &bpf) < 0) {
		pcap_close(pcap);
		return (NULL);
	}

	pcap_lookupnet(*dev_name, &local_ip, &mask, pcap_errbuf);

	return (pcap);
}

void
send_arp(eth_t *eth, u_int32_t ip)
{
	u_char pkt[sizeof(struct eth_hdr) + sizeof(struct ether_arp)];
	struct eth_hdr *ether;
	struct ether_arp *arp;
	char *buf;
	int rc;

	memset(pkt, 0, sizeof(pkt));

	memcpy(pkt, &eth_header, sizeof(eth_header));

	arp = (struct ether_arp *)(pkt + sizeof(struct eth_hdr));
	arp->ea_hdr.ar_hrd = htons(ARP_HRD_ETH);
	arp->ea_hdr.ar_pro = htons(ETH_TYPE_IP);
	arp->ea_hdr.ar_hln = ETH_ADDR_LEN;
	arp->ea_hdr.ar_pln = sizeof(ip);
	arp->ea_hdr.ar_op = htons(ARP_OP_REQUEST);

	memcpy(&arp->arp_sha, &eth_header.eth_src, sizeof(eth_header.eth_src));
	memcpy(&arp->arp_spa, &local_ip, sizeof(local_ip));
	memcpy(&arp->arp_tpa, &ip, sizeof(ip));

	eth_send(eth, (void *)pkt, sizeof(pkt));
}

int
main(int argc, char **argv)
{
	struct in_addr dummy_addr;
	u_int32_t ip_lo, ip_hi;
	char *cp, ch;
	char *dev_name;
	pcap_t *pcap;
	eth_t *dnet_eth;
	int i;

	/* Parse command line options. */
	while ((ch = getopt(argc, argv, "i:")) != -1) {
		switch (ch) {
		case 'i':
			dev_name = optarg;
			break;
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;

	/* Parse ip addresses. */
	if (argc < 1)
		usage();
	if ((cp = strchr(argv[0], '-')) != NULL) {
		*cp++ = '\0';
		ip_lo = inet_addr(argv[0]);
		if (ip_lo == INADDR_NONE) {
			fprintf(stderr, "invalid IP address: %s\n", argv[0]);
			usage();
		}
		ip_hi = inet_addr(cp);
		if (ip_hi == INADDR_NONE) {
			fprintf(stderr, "invalid IP address: %s\n", cp);
			usage();
		}
		if (ntohl(ip_lo) > ntohl(ip_hi)) {
			fprintf(stderr, "invalid address range: %s-%s\n",
			    argv[0], cp);
			usage();
		}
	} else if ((cp = strchr(argv[0], '/')) != NULL) {
		u_int32_t subnet;
		u_int32_t netmask;
		int bitmask;

		*cp++ = '\0';
		subnet = inet_addr(argv[0]);
		if (subnet == INADDR_NONE) {
			fprintf(stderr, "invalid subnet address: %s\n",
			    argv[0]);
			usage();
		}
		ntohl(subnet);
		bitmask = atoi(cp);
		if (bitmask < 1 || bitmask > 32) {
			fprintf(stderr, "invalid netmask: %s\n", cp);
			usage();
		}
		netmask = 0xffffffff << (32 - bitmask);
		ip_lo = (htonl(subnet & netmask));
		ip_hi = htonl(subnet | ~netmask);
	} else {
		ip_lo = inet_addr(argv[0]);
		if (ip_lo == INADDR_NONE) {
			fprintf(stderr, "invalid IP address: %s\n", argv[0]);
			usage();
		}
		ip_hi == ip_lo;
	}

	/* Init. pcap. */
	pcap = init_pcap(&dev_name);
	if (pcap == NULL)
		exit(1);

	/* Init. dnet. */
	dnet_eth = eth_open(dev_name);
	if (dnet_eth == NULL)
		errx("eth_open(%s) failed\n", dev_name);

	/* Init. ethernet header. */
	memset(&eth_header.eth_dst, 0xff, sizeof(eth_header.eth_dst));
	eth_get(dnet_eth, &eth_header.eth_src);
	eth_header.eth_type = htons(ETH_TYPE_ARP);

	/*
	 * Main loop.
	 *
	 * Send arp request, check for any replies, send again.
	 */
	do {
		send_arp(dnet_eth, ip_lo);
		pcap_dispatch(pcap, 0, (void *)pcap_cb, NULL);
		ip_lo = htonl(ntohl(ip_lo) + 1);
	} while (ntohl(ip_lo) <= ntohl(ip_hi));

	/* Linger for 5 seconds to pick up any late arrivals. */
	for (i = 0; i < 5000 / PCAP_TIMEO; i++)
		pcap_dispatch(pcap, 0, (void *)pcap_cb, NULL);

	eth_close(dnet_eth);
	pcap_close(pcap);

	return (0);
}
